import sys
from dataclasses import dataclass
from enum import Enum
from typing import Annotated, Literal, Optional, Union

import pytest
from serpyco_rs import ErrorItem, SchemaValidationError, Serializer
from serpyco_rs.metadata import Discriminator


@dataclass
class Foo:
    val: int
    type: Literal['foo'] = 'foo'


@dataclass
class Bar:
    type: Literal['bar']
    val: str


@dataclass
class Buz:
    type: int
    val: str


@dataclass
class Buzz:
    val: str


@dataclass
class Base:
    childs: list[Annotated[Union[Foo, Bar], Discriminator('type')]]


def test_tagged_union():
    serializer = Serializer(Base)
    val = Base(
        childs=[
            Foo(val=1),
            Bar(type='bar', val='12'),
        ]
    )
    raw = {'childs': [{'type': 'foo', 'val': 1}, {'type': 'bar', 'val': '12'}]}
    assert serializer.dump(val) == raw
    assert serializer.load(raw) == val


def test_tagged_union__invalid_discriminator_type():
    @dataclass
    class Inner:
        field: Annotated[Union[Foo, Bar, Buz], Discriminator('type')]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(Inner)

    assert exc_info.type is RuntimeError
    assert exc_info.value.args[0] == (
        'Type <class \'tests.test_union.Buz\'> has invalid discriminator field "type" with type "<class \'int\'>". '
        'Discriminator supports Literal[<str>], Literal[Enum] with str values.'
    )


def test_tagged_union__union_arg_has_no_discriminator_field():
    @dataclass
    class Inner:
        field: Annotated[Union[Foo, Bar, Buzz], Discriminator('type')]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(Inner)

    assert exc_info.type is RuntimeError
    assert exc_info.value.args[0] == 'Type <class \'tests.test_union.Buzz\'> does not have discriminator field "type"'


def test_tagged_union__unsupported_types():
    @dataclass
    class Inner:
        field: Annotated[Union[int, str], Discriminator('type')]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(Inner)

    assert exc_info.type is RuntimeError
    assert exc_info.value.args[0] == (
        'Unions supported only for dataclasses or attrs. Provided: typing.Union[int, str]'
    )


@dataclass
class A:
    type: Literal['A']
    params: int
    children: Optional[list['ComponentsT']] = None


@dataclass
class B:
    type: Literal['B']
    params: str
    children: Optional[list['ComponentsT']] = None


ComponentsT = Annotated[Union[A, B], Discriminator('type')]


@pytest.mark.skipif(sys.version_info < (3, 11), reason='requires python3.11 or higher')
def test_tagged_union__with_forward_refs():
    serializer = Serializer(ComponentsT, omit_none=True)
    data: ComponentsT = A(
        type='A',
        params=123,
        children=[
            A(
                type='A',
                params=1234,
            ),
            B(
                type='B',
                params='foo',
                children=[
                    B(
                        type='B',
                        params='bar',
                    )
                ],
            ),
        ],
    )
    raw_data = {
        'type': 'A',
        'params': 123,
        'children': [
            {'params': 1234, 'type': 'A'},
            {'children': [{'params': 'bar', 'type': 'B'}], 'params': 'foo', 'type': 'B'},
        ],
    }
    assert serializer.dump(data) == raw_data
    assert serializer.load(raw_data) == data


def test_tagged_union__camel_case_filed_format():
    @dataclass
    class A:
        discriminator_with_multiple_words: Literal['A']
        a: int

    @dataclass
    class B:
        discriminator_with_multiple_words: Literal['B']
        b: str

    raw_obj = [
        {
            'discriminatorWithMultipleWords': 'A',
            'a': 128,
        },
        {
            'discriminatorWithMultipleWords': 'B',
            'b': 'foo',
        },
    ]
    obj = [
        A(
            discriminator_with_multiple_words='A',
            a=128,
        ),
        B(
            discriminator_with_multiple_words='B',
            b='foo',
        ),
    ]

    serializer = Serializer(
        list[Annotated[Union[A, B], Discriminator('discriminator_with_multiple_words')]],
        camelcase_fields=True,
    )

    assert serializer.dump(obj) == raw_obj
    assert serializer.load(raw_obj) == obj


def test_tagged_union__enum_tag():
    class Type(Enum):
        A = 'A'
        B = 'B'

    @dataclass
    class A:
        type: Literal[Type.A]
        a: int

    @dataclass
    class B:
        type: Literal[Type.B]
        b: str

    raw_obj = [
        {
            'type': 'A',
            'a': 128,
        },
        {
            'type': 'B',
            'b': 'foo',
        },
    ]
    obj = [
        A(
            type=Type.A,
            a=128,
        ),
        B(
            type=Type.B,
            b='foo',
        ),
    ]

    serializer = Serializer(
        list[Annotated[Union[A, B], Discriminator('type')]],
        camelcase_fields=True,
    )

    assert serializer.dump(obj) == raw_obj
    assert serializer.load(raw_obj) == obj


def test_union_simple_types():
    serializer = Serializer(Union[int, str])
    assert serializer.dump(123) == 123
    assert serializer.load(123) == 123
    assert serializer.dump('123') == '123'
    assert serializer.load('123') == '123'


def test_load_union():
    @dataclass
    class Foo:
        val: int

    @dataclass
    class Bar:
        val: str

    serializer = Serializer(Union[Foo, Bar])

    assert serializer.load({'val': 123}) == Foo(val=123)
    assert serializer.load({'val': '123'}) == Bar(val='123')


def test_load_optional_union():
    @dataclass
    class Foo:
        val: int

    serializer = Serializer(Union[Foo, list[Foo], None])

    assert serializer.load({'val': 123}) == Foo(val=123)
    assert serializer.load([{'val': 1}, {'val': 2}]) == [Foo(val=1), Foo(val=2)]
    assert serializer.load(None) is None


def test_load_union_simple_types__invalid_type():
    serializer = Serializer(Union[int, str])
    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load(123.0)

    assert exc_info.value.errors == [
        ErrorItem(message='123.0 is not of type "typing.Union[int, str]"', instance_path='')
    ]



from dataclasses import dataclass

from serpyco_rs import Serializer


type Json = None | str | int | float | bool | list[Json] | dict[str, Json]


def test_to_json_schema_json_type():
    ser = Serializer(Json)
    ref_name = 'None | str | int | float | bool | list[Json] | dict[str, Json]'
    assert ser.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        '$ref': f'#/components/schemas/{ref_name}',
        'components': {
            'schemas': {
                ref_name: {
                    'anyOf': [
                        {'type': 'null'},
                        {'type': 'string'},
                        {'type': 'integer', 'format': 'int64'},
                        {'type': 'number'},
                        {'type': 'boolean'},
                        {'type': 'array', 'items': {'$ref': f'#/components/schemas/{ref_name}'}},
                        {'type': 'object', 'additionalProperties': {'$ref': f'#/components/schemas/{ref_name}'}},
                    ]
                }
            }
        },
    }


def test_to_json_schema__use_ref_for_same_type():
    type StrList = list[str]

    @dataclass
    class Data:
        a: StrList
        b: StrList

    ref = 'tests.json_schema.test_convert_py312.test_to_json_schema__use_ref_for_same_type.<locals>.Data'
    serializer = Serializer(Data)
    assert serializer.get_json_schema() == {
        '$ref': f'#/components/schemas/{ref}',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                ref: {
                    'properties': {
                        'a': {'$ref': '#/components/schemas/list[str]'},
                        'b': {'$ref': '#/components/schemas/list[str]'},
                    },
                    'type': 'object',
                    'required': ['a', 'b'],
                },
                'list[str]': {
                    'items': {'type': 'string'},
                    'type': 'array',
                },
            }
        },
    }



import enum
import uuid
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from typing import Optional, Union

from serpyco_rs import Serializer

from .utils import repeat


def test_dump_simple_types(bench_or_check_refcount):
    serializer = Serializer(float)
    bench_or_check_refcount.group = 'simple_types'
    bench_or_check_refcount(repeat(lambda: serializer.dump(1)))


def test_load_simple_types(bench_or_check_refcount):
    serializer = Serializer(int)
    bench_or_check_refcount.group = 'simple_types'
    bench_or_check_refcount(repeat(lambda: serializer.load(1)))


def test_dump_optional(bench_or_check_refcount):
    serializer = Serializer(Optional[int])

    bench_or_check_refcount.group = 'optional'

    def inner():
        repeat(lambda: serializer.dump(1))
        repeat(lambda: serializer.dump(None))

    bench_or_check_refcount(inner)


def test_load_optional(bench_or_check_refcount):
    serializer = Serializer(Optional[int])

    bench_or_check_refcount.group = 'optional'

    def inner():
        repeat(lambda: serializer.load(1))
        repeat(lambda: serializer.load(None))

    bench_or_check_refcount(inner)


def test_dump_list_simple_types(bench_or_check_refcount):
    serializer = Serializer(list[int])
    bench_or_check_refcount.group = 'list'
    data = list(range(1000))
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_list_simple_types(bench_or_check_refcount):
    serializer = Serializer(list[int])
    bench_or_check_refcount.group = 'list'
    data = list(range(1000))
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_small_list_simple_types(bench_or_check_refcount):
    serializer = Serializer(list[int])
    bench_or_check_refcount.group = 'small_list'
    data = list(range(10))
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_small_list_simple_types(bench_or_check_refcount):
    serializer = Serializer(list[int])
    bench_or_check_refcount.group = 'small_list'
    data = list(range(10))
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_tuple_simple_types(bench_or_check_refcount):
    serializer = Serializer(tuple[int, str, bool])
    bench_or_check_refcount.group = 'tuple'
    bench_or_check_refcount(repeat(lambda: serializer.dump((123, 'foo', True))))


def test_load_tuple_simple_types(bench_or_check_refcount):
    serializer = Serializer(tuple[int, str, bool])
    bench_or_check_refcount.group = 'tuple'
    bench_or_check_refcount(repeat(lambda: serializer.load((123, 'foo', True))))


def test_dump_dict_simple_types(bench_or_check_refcount):
    serializer = Serializer(dict[str, int])
    bench_or_check_refcount.group = 'dict'
    data = {str(i): i for i in range(1000)}
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_dump_dict_dataclass_value(bench_or_check_refcount):
    @dataclass
    class Foo:
        foo: int

    serializer = Serializer(dict[str, Foo])
    bench_or_check_refcount.group = 'dict'
    data = {str(i): Foo(i) for i in range(12)}
    bench_or_check_refcount(repeat(lambda: serializer.dump(data), count=100))


def test_load_dict_simple_types(bench_or_check_refcount):
    serializer = Serializer(dict[str, int])
    bench_or_check_refcount.group = 'dict'
    data = {str(i): i for i in range(1000)}
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_uuid(bench_or_check_refcount):
    serializer = Serializer(uuid.UUID)
    bench_or_check_refcount.group = 'uuid'
    data = uuid.uuid4()
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_uuid(bench_or_check_refcount):
    serializer = Serializer(uuid.UUID)
    bench_or_check_refcount.group = 'uuid'
    data = str(uuid.uuid4())
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_date(bench_or_check_refcount):
    serializer = Serializer(date)
    bench_or_check_refcount.group = 'date'
    data = date.today()
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_date(bench_or_check_refcount):
    serializer = Serializer(date)
    bench_or_check_refcount.group = 'date'
    data = date.today().isoformat()
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_time(bench_or_check_refcount):
    serializer = Serializer(time)
    bench_or_check_refcount.group = 'time'
    data = datetime.now().time()
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_time(bench_or_check_refcount):
    serializer = Serializer(time)
    bench_or_check_refcount.group = 'time'
    data = datetime.now().time().isoformat()
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_datetime(bench_or_check_refcount):
    serializer = Serializer(datetime)
    bench_or_check_refcount.group = 'datetime'
    data = datetime.now()
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_datetime(bench_or_check_refcount):
    serializer = Serializer(datetime)
    bench_or_check_refcount.group = 'datetime'
    data = datetime.now().isoformat()
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_decimal(bench_or_check_refcount):
    serializer = Serializer(Decimal)
    bench_or_check_refcount.group = 'decimal'
    data = Decimal('1.3')
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_decimal(bench_or_check_refcount):
    serializer = Serializer(Decimal)
    bench_or_check_refcount.group = 'decimal'
    data = '1.3'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


class FooEunm(enum.Enum):
    foo = 'foo'
    bar = 'bar'


def test_dump_enum(bench_or_check_refcount):
    serializer = Serializer(FooEunm)
    bench_or_check_refcount.group = 'enum'
    data = FooEunm.bar
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_enum(bench_or_check_refcount):
    serializer = Serializer(FooEunm)
    bench_or_check_refcount.group = 'enum'
    data = 'foo'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


@dataclass
class FooDataclass:
    foo: int
    bar: str


def test_dump_dataclass(bench_or_check_refcount):
    serializer = Serializer(FooDataclass)
    bench_or_check_refcount.group = 'dataclass'
    data = FooDataclass(foo=1, bar='2')
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_dataclass(bench_or_check_refcount):
    serializer = Serializer(FooDataclass)
    bench_or_check_refcount.group = 'dataclass'
    data = {'foo': 1, 'bar': '2'}
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


@dataclass
class Node:
    value: str
    next: Optional['Node'] = None


@dataclass
class Root:
    head: Node


def test_dump_recursive(bench_or_check_refcount):
    serializer = Serializer(Root)
    bench_or_check_refcount.group = 'recursive'
    data = Root(
        head=Node(
            value='1',
            next=Node(value='2'),
        ),
    )
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_recursive(bench_or_check_refcount):
    serializer = Serializer(Root)
    bench_or_check_refcount.group = 'recursive'
    data = {'head': {'next': {'next': None, 'value': '2'}, 'value': '1'}}
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_union(bench_or_check_refcount):
    @dataclass
    class Foo:
        foo: int

    serializer = Serializer(Union[int, Foo])
    data = Foo(foo=1)
    bench_or_check_refcount.group = 'union'
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_union(bench_or_check_refcount):
    @dataclass
    class Foo:
        foo: int

    serializer = Serializer(Union[int, Foo])
    data = {'foo': 1}
    bench_or_check_refcount.group = 'union'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))



import sys
from dataclasses import dataclass
from datetime import datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Literal, Optional, TypedDict, Union
from uuid import UUID

import pytest
from serpyco_rs import JsonSchemaBuilder, Serializer
from serpyco_rs.metadata import Alias, CamelCase, Discriminator, Max, MaxLength, Min, MinLength, OmitNone


@pytest.fixture
def ns(request) -> str:
    return f'tests.json_schema.test_convert.{request.node.name}'


def test_to_json_schema(ns):
    class EnumCls(Enum):
        a = 'a'

    @dataclass
    class InnerData:
        """Some important entity"""

        foo_filed: str

    @dataclass
    class OtherInnerData:
        """OtherInnerData"""

        optional_filed: Optional[InnerData] = None

    class AnotherInnerData(TypedDict):
        """AnotherInnerData"""

        bar: str

    @dataclass
    class Data:
        """Docs"""

        a: Annotated[int, Min(0), Max(10)]
        """A field with bounds"""
        b: float
        c: Decimal
        d: bool
        e: Annotated[str, MinLength(1), MaxLength(5)]
        f: UUID
        g: time
        h: datetime
        i: EnumCls
        j: InnerData
        k: list[int]
        l: tuple[int, str, InnerData]
        m: dict[str, int]
        n: Any
        o: Annotated[InnerData, CamelCase]
        p: Annotated[OtherInnerData, OmitNone]
        q: AnotherInnerData
        some_filed: Annotated[str, Alias('fooFiled')]

    serializer = Serializer(Data)

    assert serializer.get_json_schema() == {
        '$ref': f'#/components/schemas/{ns}.<locals>.Data',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.Data': {
                    'description': 'Docs',
                    'properties': {
                        'a': {
                            'description': 'A field with bounds',
                            'maximum': 10,
                            'minimum': 0,
                            'type': 'integer',
                            'format': 'int64',
                        },
                        'b': {'type': 'number'},
                        'c': {
                            'oneOf': [{'type': 'string', 'format': 'decimal'}, {'type': 'number', 'format': 'decimal'}]
                        },
                        'd': {'type': 'boolean'},
                        'e': {'maxLength': 5, 'minLength': 1, 'type': 'string'},
                        'f': {'format': 'uuid', 'type': 'string'},
                        'g': {
                            'format': 'time',
                            'type': 'string',
                        },
                        'h': {
                            'format': 'date-time',
                            'type': 'string',
                        },
                        'i': {'enum': ['a'], 'type': 'string'},
                        'j': {'$ref': f'#/components/schemas/{ns}.<locals>.InnerData'},
                        'k': {'items': {'type': 'integer', 'format': 'int64'}, 'type': 'array'},
                        'l': {
                            'maxItems': 3,
                            'minItems': 3,
                            'prefixItems': [
                                {'type': 'integer', 'format': 'int64'},
                                {'type': 'string'},
                                {'$ref': f'#/components/schemas/{ns}.<locals>.InnerData'},
                            ],
                            'type': 'array',
                        },
                        'm': {'additionalProperties': {'type': 'integer', 'format': 'int64'}, 'type': 'object'},
                        'n': {},
                        'p': {'$ref': f'#/components/schemas/{ns}.<locals>.OtherInnerData'},
                        'o': {'$ref': f'#/components/schemas/{ns}.<locals>.InnerData1'},
                        'q': {'$ref': f'#/components/schemas/{ns}.<locals>.AnotherInnerData'},
                        'fooFiled': {'type': 'string'},
                    },
                    'required': [
                        'a',
                        'b',
                        'c',
                        'd',
                        'e',
                        'f',
                        'g',
                        'h',
                        'i',
                        'j',
                        'k',
                        'l',
                        'm',
                        'n',
                        'o',
                        'p',
                        'q',
                        'fooFiled',
                    ],
                    'type': 'object',
                },
                f'{ns}.<locals>.InnerData': {
                    'description': 'Some important entity',
                    'properties': {'foo_filed': {'type': 'string'}},
                    'required': ['foo_filed'],
                    'type': 'object',
                },
                f'{ns}.<locals>.InnerData2': {
                    'description': 'Some important entity',
                    'properties': {'foo_filed': {'type': 'string'}},
                    'required': ['foo_filed'],
                    'type': 'object',
                },
                f'{ns}.<locals>.InnerData1': {
                    'description': 'Some important entity',
                    'properties': {'fooFiled': {'type': 'string'}},
                    'required': ['fooFiled'],
                    'type': 'object',
                },
                f'{ns}.<locals>.OtherInnerData': {
                    'description': 'OtherInnerData',
                    'properties': {
                        'optional_filed': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': f'#/components/schemas/{ns}.<locals>.InnerData2'},
                            ]
                        }
                    },
                    'type': 'object',
                },
                f'{ns}.<locals>.AnotherInnerData': {
                    'description': 'AnotherInnerData',
                    'properties': {'bar': {'type': 'string'}},
                    'required': ['bar'],
                    'type': 'object',
                },
            },
        },
    }


@pytest.mark.skipif(sys.version_info < (3, 10), reason='New style unions available after 3.10')
def test_to_json_schema__new_union_syntax():
    @dataclass
    class Data:
        """Docs"""

        a: int | None

    serializer = Serializer(Data)

    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__new_union_syntax.<locals>.Data',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__new_union_syntax.<locals>.Data': {
                    'description': 'Docs',
                    'properties': {'a': {'anyOf': [{'type': 'null'}, {'type': 'integer', 'format': 'int64'}]}},
                    'required': ['a'],
                    'type': 'object',
                }
            },
        },
    }


@dataclass
class TreeNode:
    """Node"""

    data: str
    left: Optional['TreeNode'] = None
    right: Optional['TreeNode'] = None


def test_to_json_schema__recursive_type():
    serializer = Serializer(TreeNode)

    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        '$ref': '#/components/schemas/tests.json_schema.test_convert.TreeNode',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.TreeNode': {
                    'description': 'Node',
                    'properties': {
                        'data': {'type': 'string'},
                        'left': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': '#/components/schemas/tests.json_schema.test_convert.TreeNode'},
                            ]
                        },
                        'right': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': '#/components/schemas/tests.json_schema.test_convert.TreeNode'},
                            ]
                        },
                    },
                    'required': ['data'],
                    'type': 'object',
                }
            }
        },
    }


def test_to_json_schema__literal():
    class BarEnum(Enum):
        bar = 'bar'

    serializer = Serializer(Literal['foo', 1, BarEnum.bar])
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'enum': ['foo', 1, 'bar'],
    }


def test_to_json_schema__tagged_union():
    @dataclass
    class Foo:
        val: int
        type: Literal['foo']

    @dataclass
    class Bar:
        val: str
        type: Literal['bar'] = 'bar'

    class BuzEnum(Enum):
        buz = 'buz'

    @dataclass
    class Baz:
        val: float
        type: Literal[BuzEnum.buz] = BuzEnum.buz

    @dataclass
    class Base:
        field: Annotated[Union[Foo, Bar, Baz], Discriminator('type')]

    serializer = Serializer(Base)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Base',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Bar': {
                    'properties': {'type': {'enum': ['bar']}, 'val': {'type': 'string'}},
                    'required': ['val', 'type'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Base': {
                    'properties': {
                        'field': {
                            'discriminator': {
                                'mapping': {
                                    'bar': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Bar',
                                    'buz': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Baz',
                                    'foo': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Foo',
                                },
                                'propertyName': 'type',
                            },
                            'oneOf': [
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Foo'
                                },
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Bar'
                                },
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Baz'
                                },
                            ],
                        }
                    },
                    'required': ['field'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Baz': {
                    'properties': {'type': {'enum': ['buz']}, 'val': {'type': 'number'}},
                    'required': ['val', 'type'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Foo': {
                    'properties': {'type': {'enum': ['foo']}, 'val': {'type': 'integer', 'format': 'int64'}},
                    'required': ['val', 'type'],
                    'type': 'object',
                },
            }
        },
    }


def test_to_json_schema__union():
    @dataclass
    class Foo:
        val: int

    @dataclass
    class Bar:
        val: str

    @dataclass
    class Base:
        field: Union[Foo, Bar, str]

    serializer = Serializer(Base)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Base',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Bar': {
                    'properties': {'val': {'type': 'string'}},
                    'required': ['val'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Base': {
                    'properties': {
                        'field': {
                            'anyOf': [
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Foo'
                                },
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Bar'
                                },
                                {'type': 'string'},
                            ],
                        }
                    },
                    'required': ['field'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Foo': {
                    'properties': {'val': {'type': 'integer', 'format': 'int64'}},
                    'required': ['val'],
                    'type': 'object',
                },
            }
        },
    }


def test_to_json_schema__force_none_as_default_for_optional():
    @dataclass
    class Data:
        a: Optional[int]

    serializer = Serializer(Data, force_default_for_optional=True)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__force_none_as_default_for_optional.<locals>.Data',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__force_none_as_default_for_optional.<locals>.Data': {
                    'properties': {'a': {'anyOf': [{'type': 'null'}, {'type': 'integer', 'format': 'int64'}]}},
                    'type': 'object',
                }
            }
        },
    }


def test_to_json_schema__bytes():
    @dataclass
    class Data:
        a: bytes

    serializer = Serializer(Data)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__bytes.<locals>.Data',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__bytes.<locals>.Data': {
                    'properties': {'a': {'type': 'string', 'format': 'binary'}},
                    'type': 'object',
                    'required': ['a'],
                }
            }
        },
    }


def test_to_json_schema__no_ref_sharing_between_different_generic_instances():
    @dataclass
    class Data:
        a: list[str]
        b: list[str]

    prefix = 'tests.json_schema.test_convert.test_to_json_schema__no_ref_sharing_between_different_generic_instances'
    ref = f'{prefix}.<locals>.Data'
    serializer = Serializer(Data)
    assert serializer.get_json_schema() == {
        '$ref': f'#/components/schemas/{ref}',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                ref: {
                    'properties': {
                        'a': {
                            'items': {'type': 'string'},
                            'type': 'array',
                        },
                        'b': {
                            'items': {'type': 'string'},
                            'type': 'array',
                        },
                    },
                    'type': 'object',
                    'required': ['a', 'b'],
                }
            }
        },
    }


def test_to_json_schema__enums_have_some_namespace_repr():
    class Foo(Enum):
        a = 'a'

    class NewNamespace:
        class Foo(Enum):
            b = 'b'

    assert repr(Foo) == repr(NewNamespace.Foo)

    serializer = Serializer(Union[Foo, NewNamespace.Foo])
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'anyOf': [
            {
                'enum': [
                    'a',
                ],
                'type': 'string',
            },
            {
                'enum': [
                    'b',
                ],
                'type': 'string',
            },
        ],
    }


def test_enum_x_attrs():
    class EnumCls(Enum):
        a = 'a'
        """a docstring"""

    serializer = Serializer(EnumCls)
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'enum': ['a'],
        'type': 'string',
        'x-a': 'a docstring',
    }


def test_enum_multiply_types():
    class EnumCls(Enum):
        a = 'a'
        """a docstring"""
        b = 1
        """b docstring"""

    serializer = Serializer(EnumCls)
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'enum': ['a', 1],
        'x-a': 'a docstring',
        'x-1': 'b docstring',
    }


def test_array_length():
    serializer = Serializer(Annotated[list[int], MinLength(1), MaxLength(10)])
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'array',
        'items': {'type': 'integer', 'format': 'int64'},
        'minItems': 1,
        'maxItems': 10,
    }


def test_one_dataclass_with_different_annotations__should_generate_different_schemas():
    @dataclass
    class Foo:
        filed_name: str

    @dataclass
    class Bar:
        foo1: Foo
        foo2: Annotated[Foo, CamelCase]

    schema = Serializer(Bar)
    name_prefix = (
        'tests.json_schema.test_convert.'
        'test_one_dataclass_with_different_annotations__should_generate_different_schemas.<locals>'
    )
    assert schema.get_json_schema() == {
        '$ref': f'#/components/schemas/{name_prefix}.Bar',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{name_prefix}.Bar': {
                    'properties': {
                        'foo1': {'$ref': f'#/components/schemas/{name_prefix}.Foo'},
                        'foo2': {'$ref': f'#/components/schemas/{name_prefix}.Foo1'},
                    },
                    'required': ['foo1', 'foo2'],
                    'type': 'object',
                },
                f'{name_prefix}.Foo': {
                    'properties': {'filed_name': {'type': 'string'}},
                    'required': ['filed_name'],
                    'type': 'object',
                },
                f'{name_prefix}.Foo1': {
                    'properties': {'filedName': {'type': 'string'}},
                    'required': ['filedName'],
                    'type': 'object',
                },
            }
        },
    }


class TestJsonSchemaBuilder:
    def test_build__use_custom_ref_prefix(self):
        @dataclass
        class Data:
            a: int

        serializer = Serializer(Data)
        schema_builder = JsonSchemaBuilder(ref_prefix='#/foo/bar')
        schema = schema_builder.build(serializer)
        definitions = schema_builder.get_definitions()

        assert schema == {
            '$ref': '#/foo/bar/tests.json_schema.test_convert.TestJsonSchemaBuilder.test_build__use_custom_ref_prefix.<locals>.Data',
        }

        assert definitions == {
            'tests.json_schema.test_convert.TestJsonSchemaBuilder.test_build__use_custom_ref_prefix.<locals>.Data': {
                'properties': {'a': {'type': 'integer', 'format': 'int64'}},
                'required': ['a'],
                'type': 'object',
            }
        }

    def test_build__add_dialect_uri(self):
        serializer = Serializer(int)
        schema_builder = JsonSchemaBuilder(add_dialect_uri=True)
        schema = schema_builder.build(serializer)

        assert schema == {
            'type': 'integer',
            'format': 'int64',
            '$schema': 'https://json-schema.org/draft/2020-12/schema',
        }
        assert schema_builder.get_definitions() == {}



import sys

import pytest
from serpyco_rs._describe import _NAME_CACHE, _generate_name


collect_ignore_glob = []

if sys.version_info[:2] < (3, 11):
    collect_ignore_glob += ['*_py311.py']

if sys.version_info[:2] < (3, 12):
    collect_ignore_glob += ['*_py312.py']


@pytest.fixture(autouse=True)
def _clear_name_cache():
    # Clear the name cache before each test to avoid side effects between tests
    _NAME_CACHE.clear()
    _generate_name.cache_clear()



from dataclasses import dataclass
from typing import Annotated, Any, Optional

import pytest
from serpyco_rs import Serializer
from serpyco_rs.metadata import Alias, Flatten
from typing_extensions import TypedDict


@dataclass
class Address:
    street: str
    city: str
    country: str


@dataclass
class Person:
    name: str
    age: int
    address: Annotated[Address, Flatten]
    extra: Annotated[dict[str, Any], Flatten]


def test_flatten_serialization():
    serializer = Serializer(Person)

    address = Address(street='123 Main St', city='New York', country='USA')
    person = Person(
        name='John Doe',
        age=30,
        address=address,
        extra={'phone': '555-1234', 'email': 'john@example.com'},
    )
    raw = {
        'name': 'John Doe',
        'age': 30,
        'street': '123 Main St',
        'city': 'New York',
        'country': 'USA',
        'phone': '555-1234',
        'email': 'john@example.com',
    }

    assert serializer.dump(person) == raw
    assert serializer.load(raw) == person


def test_flatten_first_position():
    @dataclass
    class FlattenFirst:
        metadata: Annotated[dict[str, Any], Flatten]
        name: str
        age: int

    serializer = Serializer(FlattenFirst)

    obj = FlattenFirst(metadata={'version': '1.0', 'author': 'test'}, name='John', age=30)
    raw = {'name': 'John', 'age': 30, 'version': '1.0', 'author': 'test'}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_flatten_last_position():
    @dataclass
    class FlattenLast:
        name: str
        age: int
        metadata: Annotated[dict[str, Any], Flatten]

    serializer = Serializer(FlattenLast)

    obj = FlattenLast(name='Bob', age=35, metadata={'location': 'NYC', 'timezone': 'EST'})
    raw = {'name': 'Bob', 'age': 35, 'location': 'NYC', 'timezone': 'EST'}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_multiple_dict_flatten_validation():
    @dataclass
    class MultipleFlatten:
        name: str
        config: Annotated[dict[str, Any], Flatten]
        age: int
        extra: Annotated[dict[str, Any], Flatten]
        status: str

    with pytest.raises(
        RuntimeError, match=r"Multiple dict flatten fields are not allowed in <class '.+\.MultipleFlatten'>"
    ):
        Serializer(MultipleFlatten)


@dataclass
class NestedA:
    field_a1: str
    field_a2: int


@dataclass
class NestedB:
    field_b1: str
    field_b2: bool


def test_multiple_struct_flatten():
    @dataclass
    class MultipleStructFlatten:
        id: int
        struct_a: Annotated[NestedA, Flatten]
        name: str
        struct_b: Annotated[NestedB, Flatten]
        status: str

    serializer = Serializer(MultipleStructFlatten)

    obj = MultipleStructFlatten(
        id=1,
        struct_a=NestedA(field_a1='value_a1', field_a2=42),
        name='test',
        struct_b=NestedB(field_b1='value_b1', field_b2=True),
        status='ok',
    )
    raw = {
        'id': 1,
        'field_a1': 'value_a1',
        'field_a2': 42,
        'name': 'test',
        'field_b1': 'value_b1',
        'field_b2': True,
        'status': 'ok',
    }

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_empty_dict_flatten():
    @dataclass
    class EmptyDictFlatten:
        name: str
        empty: Annotated[dict[str, Any], Flatten]

    serializer = Serializer(EmptyDictFlatten)

    obj = EmptyDictFlatten(name='test', empty={})
    raw = {'name': 'test'}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_flatten_with_camelcase():
    @dataclass
    class CamelCaseStruct:
        long_field_name: str
        another_field: int

    @dataclass
    class WithCamelCase:
        user_name: str
        nested_data: Annotated[CamelCaseStruct, Flatten]

    serializer = Serializer(WithCamelCase, camelcase_fields=True)

    obj = WithCamelCase(user_name='test', nested_data=CamelCaseStruct(long_field_name='value', another_field=42))
    raw = {'userName': 'test', 'longFieldName': 'value', 'anotherField': 42}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_flatten_with_alias():
    @dataclass
    class AliasedStruct:
        internal_name: Annotated[str, Alias('external_name')]
        value: int

    @dataclass
    class WithAlias:
        id: int
        data: Annotated[AliasedStruct, Flatten]

    serializer = Serializer(WithAlias)

    obj = WithAlias(id=123, data=AliasedStruct(internal_name='test', value=456))
    raw = {'id': 123, 'external_name': 'test', 'value': 456}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_flatten_dict_with_camelcase():
    @dataclass
    class CamelCaseWithDict:
        user_name: str
        meta_data: Annotated[dict[str, Any], Flatten]

    serializer = Serializer(CamelCaseWithDict, camelcase_fields=True)

    obj = CamelCaseWithDict(user_name='test', meta_data={'api_key': 'secret', 'log_level': 'debug'})
    raw = {'userName': 'test', 'api_key': 'secret', 'log_level': 'debug'}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_flatten_with_omit_none():
    @dataclass
    class OptionalStruct:
        required_field: str
        optional_field: Optional[str] = None

    @dataclass
    class WithOmitNone:
        name: str
        data: Annotated[OptionalStruct, Flatten]

    serializer = Serializer(WithOmitNone, omit_none=True)

    obj = WithOmitNone(name='test', data=OptionalStruct(required_field='value', optional_field=None))
    raw = {'name': 'test', 'required_field': 'value'}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_struct_vs_regular_field_conflict():
    @dataclass
    class ConflictStruct:
        name: str  # This conflicts with regular field
        value: int

    @dataclass
    class WithConflict:
        name: str
        nested: Annotated[ConflictStruct, Flatten]

    with pytest.raises(
        RuntimeError, match=r"Field name conflict in <class '.+\.WithConflict'>: 'name' from flattened struct field"
    ):
        Serializer(WithConflict)


def test_struct_field_conflict_resolved_with_alias():
    @dataclass
    class ConflictStruct:
        name: Annotated[str, Alias('nested_name')]  # Resolved with alias
        value: int

    @dataclass
    class WithConflictResolved:
        name: str
        nested: Annotated[ConflictStruct, Flatten]

    serializer = Serializer(WithConflictResolved)

    obj = WithConflictResolved(name='regular_name', nested=ConflictStruct(name='nested_value', value=42))
    raw = {'name': 'regular_name', 'nested_name': 'nested_value', 'value': 42}

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_struct_flatten_fields_conflict():
    @dataclass
    class StructA:
        common_field: str
        field_a: int

    @dataclass
    class StructB:
        common_field: int  # Same name but different type
        field_b: bool

    @dataclass
    class WithStructConflict:
        id: int
        struct_a: Annotated[StructA, Flatten]
        struct_b: Annotated[StructB, Flatten]

    # This should work since we don't validate conflicts between struct flatten fields
    # The behavior is that later fields overwrite earlier ones
    serializer = Serializer(WithStructConflict)

    obj = WithStructConflict(
        id=1,
        struct_a=StructA(common_field='string_value', field_a=42),
        struct_b=StructB(common_field=123, field_b=True),
    )

    assert serializer.dump(obj) == {
        'id': 1,
        'field_a': 42,
        'field_b': True,
        'common_field': 123,  # From struct_b (int)
    }


def test_deep_nested_flatten():
    @dataclass
    class Level3:
        level3_field: str

    @dataclass
    class Level2:
        level2_field: str
        level3: Annotated[Level3, Flatten]

    @dataclass
    class Level1:
        level1_field: str
        level2: Annotated[Level2, Flatten]

    serializer = Serializer(Level1)

    obj = Level1(level1_field='L1', level2=Level2(level2_field='L2', level3=Level3(level3_field='L3')))
    raw = {
        'level1_field': 'L1',
        'level2_field': 'L2',
        'level3_field': 'L3',
    }

    assert serializer.dump(obj) == raw
    assert serializer.load(raw) == obj


def test_flatten_unsupported_type_error():
    @dataclass
    class WithBadFlatten:
        name: str
        bad_field: Annotated[str, Flatten]  # str is not a dict

    with pytest.raises(RuntimeError, match="Flatten field 'bad_field' has type 'StringType' which cannot be flattened"):
        Serializer(WithBadFlatten)


def test_typeddict_flatten_basic():
    class NestedTypedDict(TypedDict):
        nested_field1: str
        nested_field2: int

    class MainTypedDict(TypedDict):
        main_field: str
        nested: Annotated[NestedTypedDict, Flatten]

    serializer = Serializer(MainTypedDict)

    data = {'main_field': 'main_value', 'nested': {'nested_field1': 'nested_value1', 'nested_field2': 42}}
    expected = {'main_field': 'main_value', 'nested_field1': 'nested_value1', 'nested_field2': 42}

    assert serializer.dump(data) == expected
    assert serializer.load(expected) == data


def test_typeddict_flatten_with_dict():
    class WithDictFlatten(TypedDict):
        name: str
        metadata: Annotated[dict[str, Any], Flatten]

    serializer = Serializer(WithDictFlatten)

    data = {'name': 'test', 'metadata': {'key1': 'value1', 'key2': 'value2'}}
    expected = {'name': 'test', 'key1': 'value1', 'key2': 'value2'}

    assert serializer.dump(data) == expected
    assert serializer.load(expected) == data


def test_typeddict_multiple_struct_flatten():
    class StructA(TypedDict):
        field_a1: str
        field_a2: int

    class StructB(TypedDict):
        field_b1: str
        field_b2: bool

    class MultipleStructFlatten(TypedDict):
        id: int
        struct_a: Annotated[StructA, Flatten]
        name: str
        struct_b: Annotated[StructB, Flatten]

    serializer = Serializer(MultipleStructFlatten)

    data = {
        'id': 1,
        'struct_a': {'field_a1': 'value_a1', 'field_a2': 42},
        'name': 'test',
        'struct_b': {'field_b1': 'value_b1', 'field_b2': True},
    }
    expected = {
        'id': 1,
        'field_a1': 'value_a1',
        'field_a2': 42,
        'name': 'test',
        'field_b1': 'value_b1',
        'field_b2': True,
    }

    assert serializer.dump(data) == expected
    assert serializer.load(expected) == data


def test_typeddict_mixed_dataclass_and_typeddict_flatten():
    @dataclass
    class DataclassStruct:
        dc_field1: str
        dc_field2: int

    class TypedDictStruct(TypedDict):
        td_field1: str
        td_field2: bool

    class MixedFlatten(TypedDict):
        id: int
        dc_data: Annotated[DataclassStruct, Flatten]
        td_data: Annotated[TypedDictStruct, Flatten]

    serializer = Serializer(MixedFlatten)

    data = {
        'id': 1,
        'dc_data': DataclassStruct(dc_field1='dc_value', dc_field2=42),
        'td_data': {'td_field1': 'td_value', 'td_field2': True},
    }
    expected = {'id': 1, 'dc_field1': 'dc_value', 'dc_field2': 42, 'td_field1': 'td_value', 'td_field2': True}

    assert serializer.dump(data) == expected
    assert serializer.load(expected) == data



from serpyco_rs._type_utils import get_type_hints


def test_get_type_hints_multiple_generic_inheritance():
    class Foo[T, KT]:
        arr: list[T]
        key: KT

    class Bar[T]:
        beer: T

    class Baz[T, KT](Foo[T, KT], Bar[T]):
        pass

    class Baz2[T, KT](Foo[T, KT], Bar[T]):
        pass

    assert get_type_hints(Baz[int, bool]) == {'arr': list[int], 'key': bool, 'beer': int}
    assert get_type_hints(Baz2[int, bool]) == {'arr': list[int], 'key': bool, 'beer': int}


def test_get_type_hints_swapped_typevar_order():
    class Foo[T, U]:
        type: T
        info: U

    class Bar[T](Foo[str, T]):
        test: T

    assert get_type_hints(Bar[int]) == {'type': str, 'info': int, 'test': int}


def test_get_type_hints_deep_generic_inheritance():
    class Foo[T]:
        fld: list[T]

    class Bar[T](Foo[T]):
        pass

    class Baz[T](Bar[T]):
        pass

    assert get_type_hints(Baz[int]) == {
        'fld': list[int],
    }


def test_get_type_hints_deep_generic_inheritance_different_typvar_name():
    class Foo[T]:
        fld: list[T]

    class Bar[B](Foo[B]):
        pass

    class Baz[C](Bar[C]):
        pass

    assert get_type_hints(Baz[int]) == {
        'fld': list[int],
    }


def test_get_type_hints_deep_generic_inheritance_swapped_typevar_order():
    class Foo[T1, T2]:
        fld1: list[T1]
        fld2: list[T2]

    class Bar[T1, T2](Foo[T2, T1]):
        pass

    class Baz[T1, T2](Bar[T1, T2]):
        pass

    assert get_type_hints(Baz[int, str]) == {
        'fld1': list[str],
        'fld2': list[int],
    }


def test_multiple_inheritance_with_generics():
    class Mixin1[T]:
        field1: T

    class Mixin2[U]:
        field2: list[U]

    class Base[T, U](Mixin1[T], Mixin2[U]):
        pass

    class Derived[T, U](Base[T, U]):
        pass

    assert get_type_hints(Derived[int, str]) == {'field1': int, 'field2': list[str]}


def test_swapped_typevar_order():
    class Base[T, U]:
        field1: T
        field2: U

    class Swapped[A, B](Base[B, A]):
        pass

    class Final[X, Y](Swapped[X, Y]):
        pass

    assert get_type_hints(Final[str, int]) == {
        'field1': int,  # B -> Y -> int
        'field2': str,  # A -> X -> str
    }


def test_complex_types():
    class Base[T]:
        optional_field: T | None
        union_field: T | str
        tuple_field: tuple[T, int, T]

    class Middle[T](Base[T]):
        pass

    class Final[T](Middle[T]):
        pass

    assert get_type_hints(Final[bool]) == {
        'optional_field': bool | None,
        'union_field': bool | str,
        'tuple_field': tuple[bool, int, bool],
    }


def test_multiple_generic_bases():
    class Readable[T]:
        read_data: T

    class Writable[U]:
        write_data: U

    class Processable[V]:
        process_data: V

    class MultiBase[T, U, V](Readable[T], Writable[U], Processable[V]):
        pass

    class Final[T, U, V](MultiBase[T, U, V]):
        pass

    assert get_type_hints(Final[int, str, bool]) == {'read_data': int, 'write_data': str, 'process_data': bool}


def test_mixed_variance_inheritance():
    class Base[T, U, V]:
        field1: T
        field2: U
        field3: V

    class DropOne[A, B](Base[A, str, B]):  # U fixed as str
        pass

    class ReorderAndDrop[X](DropOne[int, X]):  # A fixed as int
        pass

    assert get_type_hints(ReorderAndDrop[bool]) == {
        'field1': int,
        'field2': str,
        'field3': bool,
    }



from pathlib import Path
from typing import Any

import orjson
import pytest

from .github_issue import mashumaro, serpyco_rs


serializers = {
    'serpyco_rs': serpyco_rs,
    'mashumaro': mashumaro,
}


@pytest.fixture(scope='module')
def data() -> dict[str, Any]:
    path = Path(__file__).parent / 'github_issue/data.json'
    return orjson.loads(path.read_text())


@pytest.mark.parametrize('lib', serializers.keys())
def test_dump_github_issue(bench_or_check_refcount, lib, data):
    serializer = serializers[lib]
    test_object = serializer.load(data)
    serializer.dump(test_object)  # warmup

    bench_or_check_refcount.group = 'dump github issue'
    bench_or_check_refcount.extra_info['lib'] = lib
    bench_or_check_refcount.extra_info['correct'] = serializer.load(serializer.dump(test_object)) == test_object
    bench_or_check_refcount(serializer.dump, test_object)


@pytest.mark.parametrize('lib', serializers.keys())
def test_load_github_issue(bench_or_check_refcount, lib, data):
    serializer = serializers[lib]
    test_object = serializer.load(data)  # warmup

    bench_or_check_refcount.group = 'load github issue'
    bench_or_check_refcount.extra_info['lib'] = lib
    bench_or_check_refcount.extra_info['correct'] = serializer.load(serializer.dump(test_object)) == test_object
    bench_or_check_refcount(serializer.load, data)



from typing import Annotated

from serpyco_rs._meta import Annotations
from serpyco_rs._secial_forms import unwrap_special_forms
from serpyco_rs.metadata import Min


def test_simple_pep_695():
    type Foo = int
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_generic_pep_695():
    type Foo[T] = list[T]
    assert unwrap_special_forms(Foo[int]) == (list[int], Annotations())
    assert str(unwrap_special_forms(Foo)[0]) == 'list[T]'


def test_pep_695_with_annotated():
    type AnnotatedInt = Annotated[int, Min(5)]
    result_type, result_metadata = unwrap_special_forms(AnnotatedInt)
    assert result_type == int
    assert result_metadata.get(Min) == Min(5)


def test_pep_695_generic_with_annotated():
    type AnnotatedList[T] = Annotated[list[T], Min(1)]
    result_type, result_metadata = unwrap_special_forms(AnnotatedList[str])
    assert str(result_type) == 'list[str]'
    assert result_metadata.get(Min) == Min(1)



from dataclasses import dataclass
from typing import Annotated, Any, TypedDict

from serpyco_rs import Serializer
from serpyco_rs.metadata import Flatten

from .utils import repeat


@dataclass
class Address:
    street: str
    city: str
    country: str


@dataclass
class PersonWithFlattenStruct:
    name: str
    age: int
    address: Annotated[Address, Flatten]


@dataclass
class PersonWithFlattenDict:
    name: str
    age: int
    metadata: Annotated[dict[str, Any], Flatten]


def test_dump_flatten_struct(bench_or_check_refcount):
    data = PersonWithFlattenStruct(
        name='John Doe', age=30, address=Address(street='123 Main St', city='New York', country='USA')
    )

    serializer = Serializer(PersonWithFlattenStruct)
    serializer.dump(data)  # warmup

    bench_or_check_refcount.group = 'flatten'
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_flatten_struct(bench_or_check_refcount):
    data = {'name': 'John Doe', 'age': 30, 'street': '123 Main St', 'city': 'New York', 'country': 'USA'}

    serializer = Serializer(PersonWithFlattenStruct)
    serializer.load(data)  # warmup

    bench_or_check_refcount.group = 'flatten'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_flatten_dict(bench_or_check_refcount):
    data = PersonWithFlattenDict(
        name='Jane Smith',
        age=25,
        metadata={'department': 'Engineering', 'level': 'Senior', 'projects': ['Project A', 'Project B']},
    )

    serializer = Serializer(PersonWithFlattenDict)
    serializer.dump(data)  # warmup

    bench_or_check_refcount.group = 'flatten'
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_flatten_dict(bench_or_check_refcount):
    data = {
        'name': 'Jane Smith',
        'age': 25,
        'department': 'Engineering',
        'level': 'Senior',
        'projects': ['Project A', 'Project B'],
    }

    serializer = Serializer(PersonWithFlattenDict)
    serializer.load(data)  # warmup

    bench_or_check_refcount.group = 'flatten'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


class AddressTypedDict(TypedDict):
    street: str
    city: str
    country: str


class PersonWithFlattenStructTypedDict(TypedDict):
    name: str
    age: int
    address: Annotated[AddressTypedDict, Flatten]


class PersonWithFlattenDictTypedDict(TypedDict):
    name: str
    age: int
    metadata: Annotated[dict[str, Any], Flatten]


def test_dump_flatten_struct_typeddict(bench_or_check_refcount):
    data = PersonWithFlattenStructTypedDict(
        name='John Doe', age=30, address=AddressTypedDict(street='123 Main St', city='New York', country='USA')
    )

    serializer = Serializer(PersonWithFlattenStructTypedDict)
    serializer.dump(data)  # warmup

    bench_or_check_refcount.group = 'flatten typeddict'
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_flatten_struct_typeddict(bench_or_check_refcount):
    data = {'name': 'John Doe', 'age': 30, 'street': '123 Main St', 'city': 'New York', 'country': 'USA'}

    serializer = Serializer(PersonWithFlattenStructTypedDict)
    serializer.load(data)  # warmup

    bench_or_check_refcount.group = 'flatten typeddict'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))


def test_dump_flatten_dict_typeddict(bench_or_check_refcount):
    data = PersonWithFlattenDictTypedDict(
        name='Jane Smith',
        age=25,
        metadata={'department': 'Engineering', 'level': 'Senior', 'projects': ['Project A', 'Project B']},
    )

    serializer = Serializer(PersonWithFlattenDictTypedDict)
    serializer.dump(data)  # warmup

    bench_or_check_refcount.group = 'flatten typeddict'
    bench_or_check_refcount(repeat(lambda: serializer.dump(data)))


def test_load_flatten_dict_typeddict(bench_or_check_refcount):
    data = {
        'name': 'Jane Smith',
        'age': 25,
        'department': 'Engineering',
        'level': 'Senior',
        'projects': ['Project A', 'Project B'],
    }

    serializer = Serializer(PersonWithFlattenDictTypedDict)
    serializer.load(data)  # warmup

    bench_or_check_refcount.group = 'flatten typeddict'
    bench_or_check_refcount(repeat(lambda: serializer.load(data)))



import sys
import types
import typing
from contextlib import suppress
from typing import Annotated, Any, ClassVar, Final, ForwardRef, Union, get_origin

from typing_extensions import NotRequired, ReadOnly, Required, get_args

from ._meta import Annotations


if sys.version_info >= (3, 12):
    from typing import TypeAliasType
else:
    TypeAliasType = None


def unwrap_special_forms(annotation: Any) -> tuple[Any, Annotations]:
    metadata: list[Any] = []

    while True:
        annotation, _meta = _unpack_annotated(annotation)
        if _meta:
            metadata = _meta + metadata
            continue

        if is_typealiastype(annotation):
            # unwrap bare PEP 695 type aliases like ``type Foo = int``
            annotation = annotation.__value__
            continue

        if is_newtype(annotation):
            annotation = annotation.__supertype__
            continue

        if isinstance(annotation, str):
            annotation = ForwardRef(
                annotation,
                # we will use globals from resolver_context
                module=None,
                # allow use of special forms because we don't know current context
                is_class=True,
            )
            continue

        origin = get_origin(annotation)
        if origin is not None:
            if (
                is_classvar(origin)
                or is_final(origin)
                or is_required(origin)
                or is_notrequired(origin)
                or is_readonly(origin)
            ):
                annotation = annotation.__args__[0]
            elif is_typealiastype(origin):
                # unwrap PEP 695 type aliases
                args = get_args(annotation)
                annotation = annotation.__value__

                # apply generics args
                with suppress(TypeError):
                    annotation = annotation[args]
            else:
                # origin is not None but not a type qualifier not `Annotated` (e.g. `list[int]`):
                break
        else:
            break

    return annotation, Annotations(*metadata)


def is_annotated(origin: Any) -> bool:
    return origin is Annotated


def is_typealiastype(annotation: Any) -> bool:
    return TypeAliasType is not None and isinstance(annotation, TypeAliasType)


def is_newtype(annotation: Any) -> bool:
    return hasattr(annotation, '__supertype__') and callable(annotation)


def is_classvar(origin: Any) -> bool:
    return origin is ClassVar


def is_final(origin: Any) -> bool:
    return origin is Final


def is_required(origin: Any) -> bool:
    return origin is Required


def is_notrequired(origin: Any) -> bool:
    return origin is NotRequired


def is_readonly(origin: Any) -> bool:
    return origin is ReadOnly


def is_union_type(origin: Any) -> bool:
    if origin is Union:
        return True
    # Python 3.10+ union syntax (X | Y)
    if sys.version_info >= (3, 10) and origin is types.UnionType:
        return True
    return False


def is_literal_type(annotation: Any) -> bool:
    return annotation is typing.Literal


def _unpack_annotated(annotation: Any) -> tuple[Any, list[Any]]:
    origin = get_origin(annotation)
    if is_annotated(origin):
        annotated_type = annotation.__origin__
        metadata = list(annotation.__metadata__)

        # The annotated type might be a PEP 695 type alias, so we need to recursively unpack it
        annotated_type, sub_meta = _unpack_annotated(annotated_type)
        metadata = sub_meta + metadata
        return annotated_type, metadata
    if is_typealiastype(annotation):
        value = annotation.__value__
        typ, metadata = _unpack_annotated(value)
        if metadata:
            # Having metadata means the type alias' `__value__` was an `Annotated` form
            # (or, recursively, a type alias to an `Annotated` form). It is important to check
            # for this, as we don't want to unpack other type aliases (e.g. `type MyInt = int`).
            return typ, metadata
        return annotation, []
    if is_typealiastype(origin):
        # When parameterized, PEP 695 type aliases become generic aliases
        # (e.g. with `type MyList[T] = Annotated[list[T], ...]`, `MyList[int]`
        # is a generic alias).

        value = origin.__value__

        with suppress(TypeError):
            value = value[annotation.__args__]

        typ, metadata = _unpack_annotated(value)
        if metadata:
            return typ, metadata
        return annotation, []

    return annotation, []



from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

from serpyco_rs import Serializer


T = TypeVar('T')


@dataclass
class GenericType(Generic[T]):
    value: Optional[T] = None
    path: Optional[str] = None


@dataclass
class CustomType:
    q: str
    w: int


@dataclass
class A:
    a: GenericType[bool]
    b: GenericType[CustomType]
    c: GenericType[int]
    d: GenericType[float]
    e: GenericType[str]
    f: GenericType[str]
    g: Optional[GenericType[str]] = None


def test_generics__serialization():
    serializer = Serializer(A, omit_none=True)

    a = A(
        a=GenericType(
            value=True,
            path='some_path',
        ),
        b=GenericType(
            value=CustomType(
                q='q',
                w=1,
            ),
            path='some_path',
        ),
        c=GenericType(
            value=1,
            path='some_path',
        ),
        d=GenericType(
            value=2.0,
            path='some_path',
        ),
        e=GenericType(
            value='value',
            path='some_path',
        ),
        f=GenericType(),
        g=None,
    )

    raw_a = {
        'a': {
            'value': True,
            'path': 'some_path',
        },
        'b': {
            'value': {
                'q': 'q',
                'w': 1,
            },
            'path': 'some_path',
        },
        'c': {
            'value': 1,
            'path': 'some_path',
        },
        'd': {
            'value': 2.0,
            'path': 'some_path',
        },
        'e': {
            'value': 'value',
            'path': 'some_path',
        },
        'f': {},
    }

    assert serializer.dump(a) == raw_a
    assert serializer.load(raw_a) == a


def test_generics__swagger_schema():
    serializer = Serializer(A)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.test_generics.A',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.test_generics.A': {
                    'properties': {
                        'a': {'$ref': '#/components/schemas/tests.test_generics.GenericType[bool]'},
                        'b': {
                            '$ref': '#/components/schemas/tests.test_generics.GenericType[tests.test_generics.CustomType]'
                        },
                        'c': {'$ref': '#/components/schemas/tests.test_generics.GenericType[int]'},
                        'd': {'$ref': '#/components/schemas/tests.test_generics.GenericType[float]'},
                        'e': {'$ref': '#/components/schemas/tests.test_generics.GenericType[str]'},
                        'f': {'$ref': '#/components/schemas/tests.test_generics.GenericType[str]'},
                        'g': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': '#/components/schemas/tests.test_generics.GenericType[str]'},
                            ]
                        },
                    },
                    'required': ['a', 'b', 'c', 'd', 'e', 'f'],
                    'type': 'object',
                },
                'tests.test_generics.CustomType': {
                    'properties': {'q': {'type': 'string'}, 'w': {'type': 'integer', 'format': 'int64'}},
                    'required': ['q', 'w'],
                    'type': 'object',
                },
                'tests.test_generics.GenericType[bool]': {
                    'properties': {
                        'path': {'anyOf': [{'type': 'null'}, {'type': 'string'}]},
                        'value': {'anyOf': [{'type': 'null'}, {'type': 'boolean'}]},
                    },
                    'type': 'object',
                },
                'tests.test_generics.GenericType[float]': {
                    'properties': {
                        'path': {'anyOf': [{'type': 'null'}, {'type': 'string'}]},
                        'value': {'anyOf': [{'type': 'null'}, {'type': 'number'}]},
                    },
                    'type': 'object',
                },
                'tests.test_generics.GenericType[int]': {
                    'properties': {
                        'path': {'anyOf': [{'type': 'null'}, {'type': 'string'}]},
                        'value': {'anyOf': [{'type': 'null'}, {'type': 'integer', 'format': 'int64'}]},
                    },
                    'type': 'object',
                },
                'tests.test_generics.GenericType[str]': {
                    'properties': {
                        'path': {'anyOf': [{'type': 'null'}, {'type': 'string'}]},
                        'value': {'anyOf': [{'type': 'null'}, {'type': 'string'}]},
                    },
                    'type': 'object',
                },
                'tests.test_generics.GenericType[tests.test_generics.CustomType]': {
                    'properties': {
                        'path': {'anyOf': [{'type': 'null'}, {'type': 'string'}]},
                        'value': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': '#/components/schemas/tests.test_generics.CustomType'},
                            ]
                        },
                    },
                    'type': 'object',
                },
            }
        },
    }


def test_generics_inheritance():
    @dataclass
    class Parent(Generic[T]):
        value: T

    @dataclass
    class Child(Parent[bool]):
        pass

    serializer = Serializer(Child)
    assert serializer.dump(Child(42)) == {'value': 42}



from dataclasses import dataclass
from datetime import datetime
from typing import Annotated

import pytest
from serpyco_rs import ErrorItem, SchemaValidationError, Serializer
from serpyco_rs.metadata import CustomEncoder, deserialize_with, serialize_with


@dataclass
class Foo:
    val: Annotated[str, CustomEncoder[str, str](serialize=str.upper, deserialize=str.lower)]


def test_custom_encoder():
    serializer = Serializer(Foo)
    val = Foo(val='foo')
    raw = {'val': 'FOO'}
    assert serializer.dump(val) == raw
    assert serializer.load(raw) == val


def test_serialize_with():
    serializer = Serializer(Annotated[datetime, serialize_with(lambda x: x)])
    val = datetime.now()
    assert serializer.dump(val) is val


def test_deserialize_with():
    serializer = Serializer(Annotated[datetime, deserialize_with(lambda x: x)])
    val = datetime.now()
    assert serializer.load(val) is val


def test_deserialize_with__validation_error():
    serializer = Serializer(Annotated[int, deserialize_with(int)])
    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load('foo')

    assert exc_info.value.errors == [
        ErrorItem(message="ValueError: invalid literal for int() with base 10: 'foo'", instance_path='')
    ]



"""Query params in most cases, field values are of string type"""

from dataclasses import dataclass
from enum import Enum
from typing import Annotated, Literal, TypedDict, Union

import pytest
import serpyco_rs
from multidict import MultiDict
from serpyco_rs import metadata


def test_load_query_params__flat_type__exception():
    serializer = serpyco_rs.Serializer(int)
    with pytest.raises(RuntimeError) as exc_info:
        serializer.load_query_params(MultiDict({'id': 'a'}))

    assert str(exc_info.value) == 'This type is not deserializable from query params'


def test_load_query_params__int():
    @dataclass
    class Foo:
        id: int

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1'})) == Foo(id=1)


def test_load_query_params__int__empty():
    @dataclass
    class Foo:
        id: int

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict())


def test_load_query_params__int__invalid():
    @dataclass
    class Foo:
        id: int

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict({'id': '1.1'}))


def test_load_query_params__float():
    @dataclass
    class Foo:
        id: float

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1.1'})) == Foo(id=1.1)


def test_load_query_params__float__invalid():
    @dataclass
    class Foo:
        id: float

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict({'id': '1.1.1'}))


def test_load_query_params__bool():
    @dataclass
    class Foo:
        f: bool

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'f': 't'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'T'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'true'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'True'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'TRUE'})) == Foo(f=True)

    assert serializer.load_query_params(MultiDict({'f': 'f'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'F'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'false'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'False'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'FALSE'})) == Foo(f=False)


def test_load_query_params__bool__invalid():
    @dataclass
    class Foo:
        f: bool

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict({'f': '1'}))


def test_load_query_params__list():
    @dataclass
    class Foo:
        ids: list[int]

    serializer = serpyco_rs.Serializer(Foo)
    data = MultiDict([('ids', '1'), ('ids', 2), ('ids', '3')])
    assert serializer.load_query_params(data) == Foo(ids=[1, 2, 3])


def test_load_query_params__list__invalid():
    @dataclass
    class Foo:
        ids: list[int]

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict([('ids', 1), ('ids', '2.12'), ('ids', 3)]))


def test_load_query_params__list__empty():
    @dataclass
    class Foo:
        ids: list[int]

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict())


def test_load_query_params__list__optional():
    @dataclass
    class Foo:
        ids: Union[list[int], None] = None

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict()) == Foo(ids=None)


def test_load_query_params__tuple():
    @dataclass
    class Foo:
        ids: tuple[int, bool]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict([('ids', '1'), ('ids', 'true')])) == Foo(ids=(1, True))


def test_load_query_params__tuple__invalid():
    @dataclass
    class Foo:
        ids: tuple[int, bool]

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict([('ids', '1')]))


def test_load_query_params__union():
    @dataclass
    class Foo:
        f: Union[int, bool]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'f': '1'})) == Foo(f=1)
    assert serializer.load_query_params(MultiDict({'f': 'true'})) == Foo(f=True)


def test_load_query_params__custom_encoder():
    @dataclass
    class Foo:
        id: Annotated[int, metadata.deserialize_with(lambda value: int(value.split('-')[1]))]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1-2'})) == Foo(id=2)


def test_load_query_params__custom_encoder_list():
    @dataclass
    class Foo:
        id: Annotated[list[int], metadata.deserialize_with(lambda value: [int(v.split('-')[1]) for v in value])]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict([('id', '1-2'), ('id', '3-4')])) == Foo(id=[2, 4])


class FooEnum(Enum):
    A = 'a'
    B = 2


def test_load_query_params__enum():
    @dataclass
    class Foo:
        e: FooEnum

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'e': 'a'})) == Foo(e=FooEnum.A)
    assert serializer.load_query_params(MultiDict({'e': '2'})) == Foo(e=FooEnum.B)


def test_load_query_params__literal():
    @dataclass
    class Foo:
        e: Literal['a', 2]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'e': 'a'})) == Foo(e='a')
    assert serializer.load_query_params(MultiDict({'e': '2'})) == Foo(e=2)


def test_load_query_params__typed_dict():
    class Foo(TypedDict):
        id: int
        name: str

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1', 'name': 'test'})) == Foo(id=1, name='test')


def test_load_query_params__dict():
    serializer = serpyco_rs.Serializer(dict[str, int])
    assert serializer.load_query_params(MultiDict({'id': '1'})) == {'id': 1}


def test_load_query_params__dict_of_lists():
    serializer = serpyco_rs.Serializer(dict[str, list[int]])
    assert serializer.load_query_params(MultiDict([('foo', '1'), ('foo', '2'), ('bar', '3')])) == {
        'foo': [1, 2],
        'bar': [3],
    }



from ._custom_types import CustomType
from ._json_schema import JsonSchemaBuilder
from ._main import Serializer
from .exceptions import ErrorItem, SchemaValidationError, ValidationError


__all__ = [
    'CustomType',
    'ErrorItem',
    'JsonSchemaBuilder',
    'SchemaValidationError',
    'Serializer',
    'ValidationError',
]



import re
from collections.abc import Iterator
from contextlib import contextmanager
from functools import lru_cache
from typing import Generic, Optional, Protocol, TypeVar

from attributes_doc import get_attributes_doc as _get_attributes_doc
from typing_extensions import Self


CAMELCASE_RE = re.compile(r'(?!^)_([a-zA-Z])')


def to_camelcase(s: str) -> str:
    return CAMELCASE_RE.sub(lambda m: m.group(1).upper(), s).rstrip('_')


get_attributes_doc = lru_cache(_get_attributes_doc)


_T = TypeVar('_T')


class _Stack(Generic[_T]):
    def __init__(self, init: Optional[_T] = None) -> None:
        self._stack = [init] if init is not None else []

    @contextmanager
    def push(self, val: _T) -> Iterator[None]:
        self._stack.append(val)
        yield
        self._stack.pop()

    def get(self) -> _T:
        return self._stack[-1]


class SupportsMerge(Protocol):
    def merge(self, other: Self) -> Self: ...


_T_Merge = TypeVar('_T_Merge', bound=SupportsMerge)


class _MergeStack(_Stack[_T_Merge]):
    @contextmanager
    def merge(self, other: _T_Merge) -> Iterator[_T_Merge]:
        current = self.get()
        with self.push(current.merge(other)):
            yield self.get()



from ._convert import JsonSchemaBuilder, get_json_schema
from ._entities import ArrayType, Boolean, IntegerType, Null, NumberType, ObjectType, Schema, StringType


__all__ = [
    'ArrayType',
    'Boolean',
    'IntegerType',
    'JsonSchemaBuilder',
    'Null',
    'NumberType',
    'ObjectType',
    'Schema',
    'StringType',
    'get_json_schema',
]



from dataclasses import dataclass
from typing import Any, Optional

from mashumaro import DataClassDictMixin
from mashumaro.mixins.json import DataClassJSONMixin

from .base import make_test_object


@dataclass
class Nested(DataClassJSONMixin, DataClassDictMixin):
    """
    A nested type for Dataclass
    """

    name: str


@dataclass
class Dataclass(DataClassJSONMixin, DataClassDictMixin):
    """
    A Dataclass class
    """

    name: str
    value: int
    f: float
    b: bool
    nest: list[Nested]
    many: list[int]
    option: Optional[str] = None


test_object = make_test_object(Dataclass, Nested)


def load(data: dict[str, Any]) -> Dataclass:
    return Dataclass.from_dict(data)


def dump(obj: Dataclass) -> dict[str, Any]:
    return obj.to_dict()



from dataclasses import dataclass
from typing import Annotated, Any, Optional

import pytest
from serpyco_rs import Serializer
from serpyco_rs.metadata import Alias, Flatten


@pytest.fixture
def ns(request) -> str:
    return f'tests.json_schema.test_flatten.{request.node.name}'


def test_flatten_struct_json_schema(ns):
    @dataclass
    class Address:
        street: str
        city: str
        country: str

    @dataclass
    class Person:
        name: str
        age: int
        address: Annotated[Address, Flatten]

    serializer = Serializer(Person)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.Person',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.Person': {
                    'properties': {
                        'name': {'type': 'string'},
                        'age': {'type': 'integer', 'format': 'int64'},
                        'street': {'type': 'string'},
                        'city': {'type': 'string'},
                        'country': {'type': 'string'},
                    },
                    'required': ['name', 'age', 'street', 'city', 'country'],
                    'type': 'object',
                },
            },
        },
    }


def test_flatten_dict_json_schema(ns):
    @dataclass
    class PersonWithMeta:
        name: str
        age: int
        metadata: Annotated[dict[str, Any], Flatten]

    serializer = Serializer(PersonWithMeta)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.PersonWithMeta',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.PersonWithMeta': {
                    'properties': {
                        'name': {'type': 'string'},
                        'age': {'type': 'integer', 'format': 'int64'},
                    },
                    'required': ['name', 'age'],
                    'type': 'object',
                    'additionalProperties': True,
                },
            },
        },
    }


def test_multiple_flatten_structs_json_schema(ns):
    @dataclass
    class ContactInfo:
        email: str
        phone: str

    @dataclass
    class Address:
        street: str
        city: str

    @dataclass
    class PersonWithMultipleFlatten:
        name: str
        contact: Annotated[ContactInfo, Flatten]
        address: Annotated[Address, Flatten]

    serializer = Serializer(PersonWithMultipleFlatten)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.PersonWithMultipleFlatten',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.PersonWithMultipleFlatten': {
                    'properties': {
                        'name': {'type': 'string'},
                        'email': {'type': 'string'},
                        'phone': {'type': 'string'},
                        'street': {'type': 'string'},
                        'city': {'type': 'string'},
                    },
                    'required': ['name', 'email', 'phone', 'street', 'city'],
                    'type': 'object',
                },
            },
        },
    }


def test_flatten_with_alias_json_schema(ns):
    @dataclass
    class UserData:
        internal_field: Annotated[str, Alias('externalField')]
        value: int

    @dataclass
    class PersonWithAlias:
        name: str
        user_data: Annotated[UserData, Flatten]

    serializer = Serializer(PersonWithAlias)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.PersonWithAlias',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.PersonWithAlias': {
                    'properties': {
                        'name': {'type': 'string'},
                        'externalField': {'type': 'string'},
                        'value': {'type': 'integer', 'format': 'int64'},
                    },
                    'required': ['name', 'externalField', 'value'],
                    'type': 'object',
                },
            },
        },
    }


def test_flatten_with_camelcase_json_schema(ns):
    @dataclass
    class NestedData:
        long_field_name: str
        another_field: int

    @dataclass
    class PersonWithCamelCase:
        user_name: str
        nested_data: Annotated[NestedData, Flatten]

    serializer = Serializer(PersonWithCamelCase, camelcase_fields=True)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.PersonWithCamelCase',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.PersonWithCamelCase': {
                    'properties': {
                        'userName': {'type': 'string'},
                        'longFieldName': {'type': 'string'},
                        'anotherField': {'type': 'integer', 'format': 'int64'},
                    },
                    'required': ['userName', 'longFieldName', 'anotherField'],
                    'type': 'object',
                },
            },
        },
    }


def test_nested_flatten_json_schema(ns):
    @dataclass
    class Level3:
        level3_field: str

    @dataclass
    class Level2:
        level2_field: str
        level3: Annotated[Level3, Flatten]

    @dataclass
    class Level1:
        level1_field: str
        level2: Annotated[Level2, Flatten]

    serializer = Serializer(Level1)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.Level1',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.Level1': {
                    'properties': {
                        'level1_field': {'type': 'string'},
                        'level2_field': {'type': 'string'},
                        'level3_field': {'type': 'string'},
                    },
                    'required': ['level1_field', 'level2_field', 'level3_field'],
                    'type': 'object',
                },
            },
        },
    }


def test_flatten_with_optional_fields_json_schema(ns):
    @dataclass
    class OptionalData:
        required_field: str
        optional_field: Optional[str] = None

    @dataclass
    class PersonWithOptional:
        name: str
        data: Annotated[OptionalData, Flatten]

    serializer = Serializer(PersonWithOptional)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.PersonWithOptional',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.PersonWithOptional': {
                    'properties': {
                        'name': {'type': 'string'},
                        'required_field': {'type': 'string'},
                        'optional_field': {
                            'anyOf': [
                                {'type': 'null'},
                                {'type': 'string'},
                            ]
                        },
                    },
                    'required': ['name', 'required_field'],
                    'type': 'object',
                },
            },
        },
    }


def test_flatten_dict_with_specific_value_type_json_schema(ns):
    """Test that flatten dict fields include specific value type schema"""

    @dataclass
    class PersonWithTypedMeta:
        name: str
        age: int
        metadata: Annotated[dict[str, int], Flatten]

    serializer = Serializer(PersonWithTypedMeta)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.PersonWithTypedMeta',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.PersonWithTypedMeta': {
                    'properties': {
                        'name': {'type': 'string'},
                        'age': {'type': 'integer', 'format': 'int64'},
                    },
                    'required': ['name', 'age'],
                    'type': 'object',
                    'additionalProperties': {'type': 'integer', 'format': 'int64'},
                },
            },
        },
    }


def test_flatten_mixed_struct_and_dict_json_schema(ns):
    @dataclass
    class BaseInfo:
        info_field: str

    @dataclass
    class MixedFlatten:
        name: str
        base: Annotated[BaseInfo, Flatten]
        extra: Annotated[dict[str, Any], Flatten]

    serializer = Serializer(MixedFlatten)
    schema = serializer.get_json_schema()

    assert schema == {
        '$ref': f'#/components/schemas/{ns}.<locals>.MixedFlatten',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{ns}.<locals>.MixedFlatten': {
                    'properties': {
                        'name': {'type': 'string'},
                        'info_field': {'type': 'string'},
                    },
                    'required': ['name', 'info_field'],
                    'type': 'object',
                    'additionalProperties': True,
                },
            },
        },
    }



import dataclasses
import sys
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum, IntEnum
from functools import cache
from typing import (
    Any,
    ForwardRef,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
    get_origin,
    overload,
)
from uuid import UUID

from typing_extensions import (
    TypeGuard,
    assert_never,
    get_args,
    is_typeddict,
)

from ._custom_types import CustomType as CustomTypeMeta
from ._impl import (
    NOT_SET,
    AnyType,
    ArrayType,
    BaseType,
    BooleanType,
    BytesType,
    ContainerBaseType,
    CustomEncoder,
    CustomType,
    DateTimeType,
    DateType,
    DecimalType,
    DefaultValue,
    DictionaryType,
    DiscriminatedUnionType,
    EntityField,
    EntityType,
    EnumType,
    FloatType,
    IntegerType,
    LiteralType,
    NoneType,
    OptionalType,
    RecursionHolder,
    StringType,
    TimeType,
    TupleType,
    TypedDictType,
    UnionType,
    UUIDType,
)
from ._meta import Annotations, ResolverContext
from ._secial_forms import is_union_type, unwrap_special_forms
from ._type_utils import get_type_hints  # type: ignore[attr-defined]
from ._utils import _MergeStack, _Stack, get_attributes_doc, to_camelcase
from .metadata import (
    Alias,
    Discriminator,
    FieldFormat,
    Format,
    KeepNone,
    Max,
    MaxLength,
    Min,
    MinLength,
    NoFormat,
    NoneAsDefaultForOptional,
    NoneFormat,
    OmitNone,
    _Flatten,
)


FrozenInstanceErrors: tuple[type[Exception], ...] = (dataclasses.FrozenInstanceError,)

try:
    import attr
    from attr.exceptions import FrozenInstanceError

    FrozenInstanceErrors += (FrozenInstanceError,)
except ImportError:  # pragma: no cover
    attr = None  # type: ignore


_NoneType = type(None)

_T = TypeVar('_T')


class _TypeResolver:
    resolver_context: _Stack[ResolverContext]
    annotations: _MergeStack[Annotations]
    custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]]

    def __init__(
        self,
        globals: dict[str, Any],
        custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]] = None,
    ):
        self.resolver_context = _Stack(ResolverContext(globals=globals))
        self.custom_type_resolver = custom_type_resolver
        self.annotations = _MergeStack(Annotations(NoFormat, KeepNone, KeepNone))

    def resolve(self, t: Any) -> BaseType:
        type_info, _ = self._resolve_with_meta(t)
        return type_info

    def _resolve_with_meta(self, t: Any) -> tuple[BaseType, Annotations]:
        t, metadata = unwrap_special_forms(t)
        with self.annotations.merge(other=metadata):
            context = self.resolver_context.get()
            metadata = self.annotations.get()

            # Try find type in state
            ref = f'{t!r}::{id(t)}::{metadata.key}'
            if context.has_cached_type(ref):
                type_info = context.get_cached_type(ref)

                # for non container types avoid wrap type_info
                if type_info and not isinstance(type_info, ContainerBaseType):
                    return type_info, metadata

                if isinstance(type_info, ContainerBaseType):
                    type_info.set_usages(context.get_usage_count(ref))

                return (
                    RecursionHolder(
                        name=_generate_name(t, ref),
                        state_key=ref,
                        meta=context,
                        custom_encoder=None,
                    ),
                    metadata,
                )

            context.cache_type(ref, None)

            if isinstance(t, ForwardRef):
                t = _evaluate_forwardref(t, self.resolver_context.get())
                return self._resolve_with_meta(t)

            type_info = self._resolve_type(t, ref=ref)
            if isinstance(type_info, ContainerBaseType):
                type_info.set_usages(context.get_usage_count(ref))
            context.cache_type(ref, type_info)

            return type_info, metadata

    def _resolve_type(self, t: Any, ref: str) -> BaseType:
        annotations = self.annotations.get()
        custom_encoder = annotations.get(CustomEncoder)
        name = _generate_name(t, ref)

        if self.custom_type_resolver and (custom_type := self.custom_type_resolver(t)):
            if custom_encoder is None:
                custom_encoder = CustomEncoder(
                    serialize=custom_type.serialize,
                    deserialize=custom_type.deserialize,
                )
            return CustomType(custom_encoder=custom_encoder, json_schema=custom_type.get_json_schema())

        if t is Any:
            return AnyType(custom_encoder=custom_encoder)

        if res := self._match_simple_types(t):
            return res

        origin = get_origin(t) or t
        args = get_args(t)

        # generics types
        if origin in {Sequence, list}:
            min_length_meta = annotations.get(MinLength)
            max_length_meta = annotations.get(MaxLength)

            return ArrayType(
                item_type=(self.resolve(args[0]) if args else AnyType(custom_encoder=None)),
                ref_name=name,
                min_length=min_length_meta.value if min_length_meta else None,
                max_length=max_length_meta.value if max_length_meta else None,
                custom_encoder=custom_encoder,
            )

        if origin in {Mapping, dict}:
            return DictionaryType(
                key_type=(self.resolve(args[0]) if args else AnyType(custom_encoder=None)),
                value_type=(self.resolve(args[1]) if args else AnyType(custom_encoder=None)),
                omit_none=annotations.get(NoneFormat, KeepNone).omit,
                custom_encoder=custom_encoder,
            )

        if origin is tuple:
            if not args or Ellipsis in args:
                raise RuntimeError('Variable length tuples are not supported')
            return TupleType(
                item_types=[self.resolve(arg) for arg in args],
                ref_name=name,
                custom_encoder=custom_encoder,
            )

        if dataclasses.is_dataclass(origin) or _is_attrs(origin) or is_typeddict(origin):
            entity_type = self._describe_entity(
                t=t,
                name=_generate_name(t, ref),
                custom_encoder=custom_encoder,
            )
            return entity_type

        if _is_literal_type(origin):
            if args and _is_supported_literal_args(args):
                return LiteralType(args=list(args), custom_encoder=custom_encoder)
            raise RuntimeError('Supported only Literal[str | int, ...]')

        if is_union_type(origin):
            if _NoneType in args and len(args) == 2:
                new_arg = next(arg for arg in args if arg is not _NoneType)
                return OptionalType(
                    inner=self.resolve(new_arg),
                    custom_encoder=None,
                )

            discriminator = annotations.get(Discriminator)
            if not discriminator:
                return UnionType(
                    item_types=[self.resolve(arg) for arg in args],
                    ref_name=name,
                    custom_encoder=custom_encoder,
                )

            if not all(
                _applies_to_type_or_origin(arg, dataclasses.is_dataclass) or _applies_to_type_or_origin(arg, _is_attrs)
                for arg in args
            ):
                raise RuntimeError(f'Unions supported only for dataclasses or attrs. Provided: {t}')

            with self.resolver_context.push(
                dataclasses.replace(self.resolver_context.get(), discriminator_field=discriminator.name)
            ):
                return DiscriminatedUnionType(
                    item_types={
                        _get_discriminator_value(get_origin(arg) or arg, discriminator.name): self.resolve(arg)
                        for arg in args
                    },
                    ref_name=name,
                    dump_discriminator=discriminator.name,
                    load_discriminator=_apply_format(annotations.get(FieldFormat, NoFormat), discriminator.name),
                    custom_encoder=custom_encoder,
                )

        if isinstance(t, TypeVar):
            raise RuntimeError(f'Unfilled TypeVar: {t}')

        raise RuntimeError(f'Unknown type {t!r}')

    def _match_simple_types(self, t: Any) -> Optional[BaseType]:
        if not isinstance(t, type):
            return None

        annotations = self.annotations.get()
        custom_encoder = annotations.get(CustomEncoder)

        simple_type_mapping: Mapping[type, type[BaseType]] = {
            bytes: BytesType,
            bool: BooleanType,
            date: DateType,
            time: TimeType,
            datetime: DateTimeType,
            UUID: UUIDType,
            _NoneType: NoneType,
        }

        if simple := simple_type_mapping.get(t):
            return simple(custom_encoder=custom_encoder)

        number_type_mapping: Mapping[type, type[Union[IntegerType, FloatType]]] = {
            int: IntegerType,
            float: FloatType,
        }

        if number_type := number_type_mapping.get(t):
            min_meta = annotations.get(Min)
            max_meta = annotations.get(Max)
            return number_type(
                min=cast(Any, min_meta.value) if min_meta else None,
                max=cast(Any, max_meta.value) if max_meta else None,
                custom_encoder=custom_encoder,
            )

        if t is Decimal:
            min_meta = annotations.get(Min)
            max_meta = annotations.get(Max)
            return DecimalType(
                min=min_meta.value if min_meta else None,
                max=max_meta.value if max_meta else None,
                custom_encoder=custom_encoder,
            )

        if t is str:
            min_length_meta = annotations.get(MinLength)
            max_length_meta = annotations.get(MaxLength)

            return StringType(
                min_length=min_length_meta.value if min_length_meta else None,
                max_length=max_length_meta.value if max_length_meta else None,
                custom_encoder=custom_encoder,
            )

        if issubclass(t, (Enum, IntEnum)):
            return EnumType(cls=t, items=list(t), custom_encoder=custom_encoder)

        return None

    def _validate_flatten_fields(self, fields: list[EntityField], t: Any) -> None:
        dict_flatten_fields = []
        struct_flatten_fields = []
        regular_field_keys = set()

        for field in fields:
            if field.is_flattened:
                if isinstance(field.field_type, DictionaryType):
                    dict_flatten_fields.append(field)
                elif isinstance(field.field_type, (EntityType, TypedDictType)):
                    struct_flatten_fields.append(field)
                else:
                    raise RuntimeError(
                        f"Flatten field '{field.name}' has type '{type(field.field_type).__name__}' "
                        f'which cannot be flattened. Only dataclasses, TypedDict, and dict types '
                        f'can be used with Flatten annotation.'
                    )
            else:
                regular_field_keys.add(field.dict_key)

        if len(dict_flatten_fields) > 1:
            field_names = [field.name for field in dict_flatten_fields]
            raise RuntimeError(
                f'Multiple dict flatten fields are not allowed in {t!r}. '
                f'Found flatten dict fields: {field_names}. '
                f'Use only one Annotated[dict[str, Any], Flatten] field per dataclass.'
            )

        for struct_field in struct_flatten_fields:
            struct_type = cast(Union[EntityType, TypedDictType], struct_field.field_type)
            for nested_field in struct_type.fields:
                if nested_field.dict_key in regular_field_keys:
                    raise RuntimeError(
                        f"Field name conflict in {t!r}: '{nested_field.dict_key}' from "
                        f"flattened struct field '{struct_field.name}' conflicts with regular field. "
                        f'Use Alias annotation to rename one of the conflicting fields.'
                    )

    def _compute_used_keys(self, fields: list[EntityField]) -> set[str]:
        has_dict_flatten = any(field.is_dict_flatten for field in fields)
        if not has_dict_flatten:
            return set()

        used_keys: set[str] = set()

        for field in fields:
            if field.is_flattened and isinstance(field.field_type, (EntityType, TypedDictType)):
                struct_type = field.field_type
                for nested_field in struct_type.fields:
                    used_keys.add(nested_field.dict_key)
            else:
                used_keys.add(field.dict_key)

        return used_keys

    def _describe_entity(
        self,
        t: Any,
        name: str,
        custom_encoder: Optional[CustomEncoder[Any, Any]],
    ) -> Union[EntityType, TypedDictType]:
        # PEP-484: Replace all unfilled type parameters with Any
        if get_origin(t) is None and getattr(t, '__parameters__', None):
            t = t[(Any,) * len(t.__parameters__)]

        origin = get_origin(t) or t
        docs = get_attributes_doc(t)
        try:
            types = get_type_hints(t, include_extras=True)
        except Exception:  # pylint: disable=broad-except
            types = {}

        discriminator_field = self.resolver_context.get().discriminator_field

        cls_annotations = self.annotations.get()

        with self.resolver_context.push(
            dataclasses.replace(self.resolver_context.get(), globals=_get_globals(t), discriminator_field=None)
        ):
            fields = []
            for field in _get_entity_fields(origin):
                type_ = types.get(field.name, field.type)

                if isinstance(type_, str):
                    type_ = ForwardRef(
                        type_,
                        module=t.__module__,
                        is_class=True,
                    )
                field_type, field_metadata = self._resolve_with_meta(type_)
                with self.annotations.merge(field_metadata) as field_annotations:
                    alias = field_annotations.get(Alias)
                    none_as_default_for_optional = field_annotations.get(NoneAsDefaultForOptional)
                    flatten = field_annotations.get(_Flatten)

                    is_discriminator_field = field.name == discriminator_field
                    required = (
                        not (field.default != NOT_SET or field.default_factory != NOT_SET) or is_discriminator_field
                    )

                    default = field.default
                    if (
                        isinstance(field_type, OptionalType)
                        and required
                        and none_as_default_for_optional
                        and none_as_default_for_optional.use
                    ):
                        default = DefaultValue.some(None)
                        required = False

                    is_flattened = flatten is not None
                    is_dict_flatten = is_flattened and isinstance(field_type, DictionaryType)

                    fields.append(
                        EntityField(
                            name=field.name,
                            dict_key=(
                                alias.value if alias else _apply_format(cls_annotations.get(FieldFormat), field.name)
                            ),
                            doc=docs.get(field.name),
                            field_type=field_type,
                            default=default,
                            default_factory=field.default_factory,
                            is_discriminator_field=is_discriminator_field,
                            required=required,
                            is_flattened=is_flattened,
                            is_dict_flatten=is_dict_flatten,
                        )
                    )

            self._validate_flatten_fields(fields, t)
            used_keys = self._compute_used_keys(fields)

            if is_typeddict(origin):
                return TypedDictType(
                    name=name,
                    fields=fields,
                    omit_none=cls_annotations.get(NoneFormat) is OmitNone,
                    doc=t.__doc__,
                    custom_encoder=custom_encoder,
                    used_keys=used_keys,
                )

            return EntityType(
                cls=origin,
                name=name,
                fields=fields,
                omit_none=cls_annotations.get(NoneFormat) is OmitNone,
                is_frozen=_is_frozen_dataclass(origin, fields[0]) if fields else False,
                doc=_get_dataclass_doc(t),
                custom_encoder=custom_encoder,
                used_keys=used_keys,
            )


def describe_type(
    t: Any,
    meta: Optional[ResolverContext] = None,
    custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]] = None,
) -> BaseType:
    return _TypeResolver(_get_globals(t), custom_type_resolver).resolve(t)


@dataclasses.dataclass
class _Field(Generic[_T]):
    name: str
    type: Union[type[_T], str, Any]
    default: Union[DefaultValue[_T], DefaultValue[None]] = NOT_SET
    default_factory: Union[DefaultValue[Callable[[], _T]], DefaultValue[None]] = NOT_SET


def _get_entity_fields(t: Any) -> Sequence[_Field[Any]]:
    if dataclasses.is_dataclass(t):
        return [
            _Field(
                name=f.name,
                type=f.type,
                default=(DefaultValue.some(f.default) if f.default is not dataclasses.MISSING else NOT_SET),
                default_factory=(
                    DefaultValue.some(f.default_factory) if f.default_factory is not dataclasses.MISSING else NOT_SET
                ),
            )
            for f in dataclasses.fields(t)
        ]
    if is_typeddict(t):
        return [
            _Field(
                name=field_name,
                type=field_type,
                default=NOT_SET if _is_required_in_typeddict(t, field_name) else DefaultValue.some(None),
                default_factory=NOT_SET,
            )
            for field_name, field_type in t.__annotations__.items()
        ]
    if _is_attrs(t):
        assert attr
        return [
            _Field(
                name=f.name,
                type=f.type,
                default=(
                    DefaultValue.some(f.default)
                    if (
                        f.default is not attr.NOTHING and not isinstance(f.default, attr.Factory)  # type: ignore[arg-type]
                    )
                    else NOT_SET
                ),
                default_factory=(
                    DefaultValue.some(f.default.factory)  # pyright: ignore
                    if isinstance(f.default, attr.Factory)  # type: ignore[arg-type]
                    else NOT_SET
                ),
            )
            for f in attr.fields(t)
        ]

    raise RuntimeError(f"Unsupported type '{t}'")


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: _T) -> _T: ...


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: None = None) -> Optional[_T]: ...


def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: Optional[_T] = None) -> Optional[_T]:
    return next((ann for ann in annotations if isinstance(ann, type_)), default)


def _apply_format(f: Optional[FieldFormat], value: str) -> str:
    if not f or f.format is Format.no_format:
        return value
    if f.format is Format.camel_case:
        return to_camelcase(value)
    assert_never(f.format)


_NAME_CACHE = {}


@cache
def _generate_name(cls: Any, ref: str) -> str:
    """
    Generate unique name for entity type.

    We need unique name to avoid name conflicts in generated json schema,
    when we have one entity with different Annotated metadata (like FieldFormat).
    """
    name = repr(cls).removeprefix("<class '").removesuffix("'>")
    if name not in _NAME_CACHE:
        _NAME_CACHE[name] = 0
        return name

    _NAME_CACHE[name] += 1
    return f'{name}{_NAME_CACHE[name]}'


def _get_globals(t: Any) -> dict[str, Any]:
    if t.__module__ in sys.modules:
        return sys.modules[t.__module__].__dict__.copy()
    return {}


def _evaluate_forwardref(t: ForwardRef, meta: ResolverContext) -> Any:
    return t._evaluate(meta.globals, {}, recursive_guard=frozenset[str]())


def _get_discriminator_value(t: Any, name: str) -> str:
    fields = attr.fields(t) if attr and _is_attrs(t) else dataclasses.fields(t)
    for field in fields:
        if field.name == name:
            if _is_literal_type(field.type):
                args = get_args(field.type)
                if len(args) != 1:
                    raise RuntimeError(
                        f'Type {t} has invalid discriminator field "{name}". '
                        f'Discriminator supports only Literal[...] with one argument.'
                    )
                arg = args[0]

                if isinstance(arg, Enum):
                    arg = arg.value

                if isinstance(arg, str):
                    return arg

            raise RuntimeError(
                f'Type {t} has invalid discriminator field "{name}" with type "{field.type!r}". '
                f'Discriminator supports Literal[<str>], Literal[Enum] with str values.'
            )
    raise RuntimeError(f'Type {t} does not have discriminator field "{name}"')


def _is_str_literal(t: Any) -> bool:
    if _is_literal_type(t):
        args = get_args(t)
        if args and all(isinstance(arg, str) for arg in args):
            return True
    return False


def _is_literal_type(t: Any) -> bool:
    return t is Literal or get_origin(t) is Literal


def _is_attrs(t: Any) -> bool:
    return attr is not None and attr.has(t)


def _is_required_in_typeddict(t: Any, key: str) -> bool:
    if is_typeddict(t):
        if t.__total__:
            return key not in t.__optional_keys__
        return key in t.__required_keys__
    raise RuntimeError(f'Expected TypedDict, got "{t!r}"')


def _get_dataclass_doc(cls: Any) -> Optional[str]:
    """Dataclass has automatically generated docstring, which is not very useful."""
    doc: str = cls.__doc__
    if doc is None or doc.startswith(f'{cls.__name__}('):
        return None
    return doc


def _is_frozen_dataclass(cls: Any, field: EntityField) -> bool:
    try:
        obj = object.__new__(cls)
        setattr(obj, field.name, None)
    except FrozenInstanceErrors:
        return True
    else:
        return False


def _is_supported_literal_args(args: Sequence[Any]) -> TypeGuard[list[Union[str, int, Enum]]]:
    return all(isinstance(arg, (str, int, Enum)) for arg in args)


def _applies_to_type_or_origin(t: Any, predicate: Callable[[Any], bool]) -> bool:
    if predicate(t):
        return True
    if hasattr(t, '__origin__'):
        return predicate(t.__origin__)
    return False



from typing import Annotated, ClassVar, Final, NewType, Union

from serpyco_rs._meta import Annotations
from serpyco_rs._secial_forms import unwrap_special_forms
from serpyco_rs.metadata import Alias, Max, Min
from typing_extensions import NotRequired, ReadOnly, Required


def test_bare_type():
    assert unwrap_special_forms(int) == (int, Annotations())


def test_new_type():
    Foo = NewType('Foo', int)
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_required():
    Foo = Required[int]
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_not_required():
    Foo = NotRequired[int]
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_class_var():
    Foo = ClassVar[int]
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_final():
    Foo = Final[int]
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_readonly():
    Foo = ReadOnly[int]
    assert unwrap_special_forms(Foo) == (int, Annotations())


def test_annotated_simple():
    Foo = Annotated[int, Min(5)]
    result_type, result_metadata = unwrap_special_forms(Foo)
    assert result_type == int
    assert result_metadata.get(Min) == Min(5)


def test_annotated_multiple_metadata():
    Foo = Annotated[str, Min(5), Max(10), Alias('foo')]
    result_type, result_metadata = unwrap_special_forms(Foo)
    assert result_type == str
    assert result_metadata.get(Min) == Min(5)
    assert result_metadata.get(Max) == Max(10)
    assert result_metadata.get(Alias) == Alias('foo')


def test_nested_annotated():
    Inner = Annotated[int, Min(5)]
    Outer = Annotated[Inner, Max(10)]
    result_type, result_metadata = unwrap_special_forms(Outer)
    assert result_type == int
    assert result_metadata.get(Min) == Min(5)
    assert result_metadata.get(Max) == Max(10)


def test_annotated_with_newtype():
    UserId = NewType('UserId', int)
    Foo = Annotated[UserId, Min(1)]
    result_type, result_metadata = unwrap_special_forms(Foo)
    assert result_type == int
    assert result_metadata.get(Min) == Min(1)


def test_annotated_with_required():
    Foo = Annotated[Required[int], Min(5)]
    result_type, result_metadata = unwrap_special_forms(Foo)
    assert result_type == int
    assert result_metadata.get(Min) == Min(5)


def test_complex_nested_forms():
    UserId = NewType('UserId', int)
    AnnotatedUserId = Annotated[UserId, Min(1)]
    RequiredAnnotatedUserId = Required[AnnotatedUserId]

    result_type, result_metadata = unwrap_special_forms(RequiredAnnotatedUserId)
    assert result_type == int
    assert result_metadata.get(Min) == Min(1)


def test_generic_types_not_unwrapped():
    result_type, result_metadata = unwrap_special_forms(list[int])
    assert result_type == list[int]
    assert result_metadata == Annotations()


def test_union_not_unwrapped():
    result_type, result_metadata = unwrap_special_forms(Union[int, str])
    assert result_type == Union[int, str]
    assert result_metadata == Annotations()



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import TypeVar, Union

from ._impl import CustomEncoder


@dataclass(frozen=True)
class Min:
    value: Union[int, float]


@dataclass(frozen=True)
class Max:
    value: Union[int, float]


@dataclass(frozen=True)
class MinLength:
    value: int


@dataclass(frozen=True)
class MaxLength:
    value: int


@dataclass(frozen=True)
class Discriminator:
    name: str


@dataclass(frozen=True)
class Alias:
    value: str


class Format(Enum):
    no_format = 'no_format'
    camel_case = 'camel_case'


@dataclass(frozen=True)
class FieldFormat:
    format: Format


CamelCase: FieldFormat = FieldFormat(Format.camel_case)
NoFormat: FieldFormat = FieldFormat(Format.no_format)


@dataclass(frozen=True)
class NoneFormat:
    omit: bool


KeepNone: NoneFormat = NoneFormat(False)
OmitNone: NoneFormat = NoneFormat(True)


@dataclass(frozen=True)
class NoneAsDefaultForOptional:
    use: bool


ForceDefaultForOptional: NoneAsDefaultForOptional = NoneAsDefaultForOptional(True)
KeepDefaultForOptional: NoneAsDefaultForOptional = NoneAsDefaultForOptional(False)


_I = TypeVar('_I')
_O = TypeVar('_O')


def serialize_with(func: Callable[[_I], _O]) -> CustomEncoder[_I, _O]:
    return CustomEncoder[_I, _O](serialize=func)


def deserialize_with(func: Callable[[_O], _I]) -> CustomEncoder[_I, _O]:
    return CustomEncoder[_I, _O](deserialize=func)


@dataclass(frozen=True)
class _Flatten:
    """Flatten the fields of a nested structure into the parent structure.

    Similar to serde's flatten attribute, this allows inlining fields from
    a nested dataclass or collecting additional fields in a dict.

    Examples:
        # Flatten a nested dataclass
        @dataclass
        class Person:
            name: str
            address: Annotated[Address, Flatten]

        # Collect additional fields in a dict
        @dataclass
        class FlexibleData:
            id: str
            extra: Annotated[dict[str, Any], Flatten]
    """


Flatten: _Flatten = _Flatten()



from dataclasses import dataclass
from typing import Annotated, Literal

import pytest
from serpyco_rs import Serializer
from serpyco_rs.metadata import Discriminator


type Json = str | int | float | bool | None | list[Json] | dict[str, Json]


def test_json_type():
    ser = Serializer(Json)
    assert ser.load(None) == None  # noqa: E711
    assert ser.load(True) == True  # noqa: E712
    assert ser.load(False) == False  # noqa: E712
    assert ser.load(123) == 123
    assert ser.load(123.456) == 123.456
    assert ser.load('hello') == 'hello'
    assert ser.load([1, 2, 3]) == [1, 2, 3]
    assert ser.load({'a': 1, 'b': 2}) == {'a': 1, 'b': 2}


type MyList[T] = list[T]


def test_generic_with_type_param():
    ser = Serializer(MyList[int])
    assert ser.load([1]) == [1]
    assert ser.dump([2]) == [2]


def test_generic_with_unfilled_type_param__fail():
    with pytest.raises(RuntimeError) as exc_info:
        Serializer(MyList)

    assert exc_info.match('Unfilled TypeVar: T')


@dataclass(kw_only=True)
class Foo[T]:
    type: Literal['Foo'] = 'Foo'
    children_field: list[T]


@dataclass(kw_only=True)
class Bar[T]:
    type: Literal['Bar'] = 'Bar'
    children_field: list[T]
    field: int | None = None


type Core = Annotated[Foo[Core] | Bar[Core], Discriminator('type')]


def test_serialize__new_type_var_recursive_generic_dataclass__correct():
    serializer = Serializer(Core, camelcase_fields=True, omit_none=True)

    obj = Bar(
        children_field=[
            Foo(
                children_field=[
                    Bar(
                        children_field=[],
                        field=1,
                    )
                ]
            )
        ]
    )

    data = {
        'childrenField': [
            {
                'childrenField': [
                    {
                        'childrenField': [],
                        'field': 1,
                        'type': 'Bar',
                    }
                ],
                'type': 'Foo',
            }
        ],
        'type': 'Bar',
    }

    assert serializer.dump(obj) == data
    assert serializer.load(data) == obj


def test_multiple_type_params():
    type MyDict[K, V] = dict[K, V]

    ser = Serializer(MyDict[str, int])
    data = {'key1': 1, 'key2': 2}
    assert ser.load(data) == data
    assert ser.dump(data) == data


def test_multiple_type_params_unfilled__fail():
    type MyDict[K, V] = dict[K, V]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(MyDict)

    assert exc_info.match('Unfilled TypeVar: K')


def test_nested_generics():
    type NestedList[T] = list[list[T]]
    ser = Serializer(NestedList[str])
    data = [['hello', 'world'], ['foo', 'bar']]
    assert ser.load(data) == data
    assert ser.dump(data) == data


def test_union_with_type_params():
    type Result[T, E] = T | E

    ser = Serializer(Result[str, int])
    assert ser.load('success') == 'success'
    assert ser.load(404) == 404
    assert ser.dump('success') == 'success'
    assert ser.dump(404) == 404


def test_complex_nested_structure():
    type ComplexStructure[T] = dict[str, list[T | None]]

    ser = Serializer(ComplexStructure[int])
    data = {'numbers': [1, 2, None, 3], 'more_numbers': [None, 4, 5]}
    assert ser.load(data) == data
    assert ser.dump(data) == data


def test_generic_dataclass_multiple_params():
    @dataclass
    class Container[T, U]:
        first: T
        second: U

    ser = Serializer(Container[str, int])
    obj = Container(first='hello', second=42)
    data = {'first': 'hello', 'second': 42}

    assert ser.dump(obj) == data
    assert ser.load(data) == obj


def test_recursive_type_with_param():
    type Tree[T] = dict[str, T | Tree[T]]
    ser = Serializer(Tree[int])
    data = {'value': 1, 'children': {'left': {'value': 2}, 'right': 3}}
    assert ser.load(data) == data
    assert ser.dump(data) == data
